/************************************************************************
	main.c

    USB Performance Monitor - PIC18F2550 firmware
    Copyright (C) 2010 Simon Inns

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	Email: simon.inns@gmail.com

************************************************************************/

#ifndef MAIN_C
#define MAIN_C

// Global includes
#include <htc.h>
#include <math.h>

// Local includes
#include "usb.h"
#include "HardwareProfile.h"
#include "usb_function_hid.h"
#include "genericHID.h"

// PIC 18F2550 fuse configuration:
// Config word 1 (Oscillator configuration)
// 20Mhz crystal input scaled to 48Mhz and configured for USB operation
__CONFIG(1, USBPLL & IESODIS & FCMDIS & HSPLL & CPUDIV1 & PLLDIV5);
// Config word 2
__CONFIG(2, VREGEN & PWRTDIS & BOREN & BORV20 & WDTDIS & WDTPS32K);
// Config word 3
__CONFIG(3, PBDIGITAL & LPT1DIS & MCLRDIS);
// Config word 4
__CONFIG(4, XINSTDIS & STVREN & LVPDIS & ICPORTDIS & DEBUGDIS);
// Config word 5, 6 and 7 (protection configuration)
__CONFIG(5, UNPROTECT);
__CONFIG(6, UNPROTECT);
__CONFIG(7, UNPROTECT);

// Globals for RGB LED PWM ----------------------------------------------------
int pwmCounter = 0;

int grnActualBrightness = 0;
int bluActualBrightness = 0;
int redActualBrightness = 0;

int grnTargetBrightness = 0;
int bluTargetBrightness = 0;
int redTargetBrightness = 0;

int grnFadeSpeed = 0;
int bluFadeSpeed = 0;
int redFadeSpeed = 0;

int grnFadeCounter = 0;
int bluFadeCounter = 0;
int redFadeCounter = 0;

// Globals for meter control --------------------------------------------------
int meter1ActualValue = 0;
int meter1TargetValue = 0;
int meter1DropRate = 0;
int meter1DropCounter = 0;

int meter2ActualValue = 0;
int meter2TargetValue = 0;
int meter2DropRate = 0;
int meter2DropCounter = 0;

// Globals for device settings
int ledFadeOnRateSetting = 100;
int ledFadeOffRateSetting = 200;
int meter1SensitivitySetting = 100;
int meter2SensitivitySetting = 100;

// Globals for hardware button
int buttonPressed = 0; // 0 = not pressed, 1 = pressed
int buttonDebounceCounter = 0;

// Globals for soft USB detach function
int softUsbDetach = 0; // 0 = false (attached), 1 = true (detached)

// Low priority interrupt procedure
void interrupt low_priority lpHandler(void)
{
	// Is this timer0 interrupting?
	if (TMR0IF)
	{
		// LED Control --------------------------------------------------------
		
		// Perform the PWM brightness control
		if (grnActualBrightness > pwmCounter) GRN = 0; else GRN = 1;
		if (bluActualBrightness > pwmCounter) BLU = 0; else BLU = 1;
		if (redActualBrightness > pwmCounter) RED = 0; else RED = 1;

		pwmCounter++;
		if (pwmCounter > 127) pwmCounter = 0;
		
		// Perform fading control (Fading for each colour is independent and
		// supports fading up and down as well as an immediate change
		
		// Green LED control
		if (grnTargetBrightness != grnActualBrightness)
		{
			// If fade is off then just set the brightness
			if (grnFadeSpeed == -1) grnActualBrightness = grnTargetBrightness;
			else
			{
				grnFadeCounter++;
				if (grnFadeCounter == grnFadeSpeed)
				{
					if (grnTargetBrightness > grnActualBrightness) grnActualBrightness++;
					else grnActualBrightness--;
					grnFadeCounter = 0;
				}
			}
		}	
		
		// Blue LED control
		if (bluTargetBrightness != bluActualBrightness)
		{
			// If fade is off then just set the brightness
			if (bluFadeSpeed == -1) bluActualBrightness = bluTargetBrightness;
			else
			{
				bluFadeCounter++;
				if (bluFadeCounter == bluFadeSpeed)
				{
					if (bluTargetBrightness > bluActualBrightness) bluActualBrightness++;
					else bluActualBrightness--;
					bluFadeCounter = 0;
				}
			}
		}	
		
		// Red LED control
		if (redTargetBrightness != redActualBrightness)
		{
			// If fade is off then just set the brightness
			if (redFadeSpeed == -1) redActualBrightness = redTargetBrightness;
			else
			{
				redFadeCounter++;
				if (redFadeCounter == redFadeSpeed)
				{
					if (redTargetBrightness > redActualBrightness) redActualBrightness++;
					else redActualBrightness--;
					redFadeCounter = 0;
				}
			}
		}
		
		// Meter control ------------------------------------------------------
		
		// Meter 1 (CPU)
		if ((meter1TargetValue > meter1ActualValue) || meter1DropRate == 0)
			meter1ActualValue = meter1TargetValue;
		else
		{
			meter1DropCounter++;
			if (meter1DropCounter == meter1DropRate)
			{
				meter1ActualValue-=2;
				if(meter1ActualValue < meter1TargetValue) meter1ActualValue = meter1TargetValue;
				meter1DropCounter = 0;
			}	
		}	
		
		// Meter 2 (HDD)
		if ((meter2TargetValue > meter2ActualValue) || meter2DropRate == 0)
			meter2ActualValue = meter2TargetValue;
		else
		{
			meter2DropCounter++;
			if (meter2DropCounter == meter2DropRate)
			{
				meter2ActualValue-=2;
				if(meter2ActualValue < meter2TargetValue) meter2ActualValue = meter2TargetValue;
				meter2DropCounter = 0;
			}	
		}	
		
		// Set the PWM generation for the meters
		CCPR1L = meter2ActualValue;
		CCPR2L = meter1ActualValue;
		
		// Get ready for the next interrupt -----------------------------------
		TMR0L = 255 - 98;	// Reset the timer0 counter
		TMR0IF = 0;			// Clear the timer0 interrupt flag
	}
}

// Set the RGB LED colour (input is 0-255) and fade time in milliseconds
void setRgbLedColour(unsigned char red, unsigned char green, unsigned char blue)
{
	// Here we specify the minimum intensity (if we allow the intensity to be too low
	// there aren't enough brightness levels to fade smoothly
	if (red != 0 && red < 20) red = 20;
	if (green != 0 && green < 20) green = 20;
	if (blue != 0 && blue < 20) blue = 20;
	
	// Scale the intensities to a range of 0-127 and normalise the brightness output
	float redScale = (47.0 * (float)red / 71.0) / 2.0;
	float greenScale = ((float)green) / 2.0;
	float blueScale = (47.0 * (float)blue / 64.0) / 2.0;
	
	// If the value are the same as currently set, just return
	if ( ((int)redScale == redTargetBrightness) &&
		((int)blueScale == bluTargetBrightness) &&
		((int)greenScale == grnTargetBrightness)) return;
	
	int fadeTime;
	if (red == 0 && blue == 0 && green ==0) fadeTime = ledFadeOffRateSetting;
	else fadeTime = ledFadeOnRateSetting;
	
	// Calculate the difference between the current colour and the new colour
	// this ensures the fade speed is the same no matter the difference in brightness levels
	// between the three colours
	float redFadeSpeedFloat =
		1 / (((float)redTargetBrightness - redScale)   / (((float)fadeTime / 130.0) * 1000.0));
	float grnFadeSpeedFloat =
		1 / (((float)grnTargetBrightness - greenScale) / (((float)fadeTime / 130.0) * 1000.0));
	float bluFadeSpeedFloat =
		1 / (((float)bluTargetBrightness - blueScale)  / (((float)fadeTime / 130.0) * 1000.0));
	
	// Set the fade speed as an integer value
	redFadeSpeed = (int)redFadeSpeedFloat;
	grnFadeSpeed = (int)grnFadeSpeedFloat;
	bluFadeSpeed = (int)bluFadeSpeedFloat;
	
	// Ensure that the fade speed increment is always positive
	if (redFadeSpeed < 0) redFadeSpeed = -redFadeSpeed;
	if (grnFadeSpeed < 0) grnFadeSpeed = -grnFadeSpeed;
	if (bluFadeSpeed < 0) bluFadeSpeed = -bluFadeSpeed;
	
	// If the fade speed is too fast we set it to -1 which tells the
	// interrupt routine to change brightness immediately
	if (redFadeSpeed == 0) redFadeSpeed = -1;
	if (grnFadeSpeed == 0) grnFadeSpeed = -1;
	if (bluFadeSpeed == 0) bluFadeSpeed = -1;
	
	// Set the target brightness levels
	redTargetBrightness = (int)redScale;
	grnTargetBrightness = (int)greenScale;
	bluTargetBrightness = (int)blueScale;
	
	redFadeCounter = 0;
	grnFadeCounter = 0;
	bluFadeCounter = 0;
}

// Set the meter values (range 0-100%) and the meter sensitivity (0-100%)
void setMeter(unsigned char meter1Percent, unsigned char meter2Percent,
	 unsigned char meter1Sensitivity, unsigned char meter2Sensitivity)
{
	double meter1;
	double meter2;
	double meter1position;
	double meter2position;

	// Set the meter drop-out rates ('sensitivity')
	if (meter1Sensitivity > 100) meter1Sensitivity = 100;
	if (meter2Sensitivity > 100) meter2Sensitivity = 100;
	meter1DropRate = (100 - meter1Sensitivity) / 2;
	meter2DropRate = (100 - meter2Sensitivity) / 2;

	// Set meter 1 (CPU) %
	if (meter1Percent > 100) meter1Percent = 100;
	meter1 = (double)meter1Percent;

	// Use a polynomial to calculate the real 0-255 based on the logarithmic scale
	meter1position = (255.0/309.0) * (0.0117 * pow(meter1, 2.0) + 1.8663 * meter1 + 3.1713);

	// Set the PWM duty-cycle
	meter1TargetValue = (unsigned char)meter1position;

	// Set meter 2 (HDD) %
	if (meter2Percent > 100) meter2Percent = 100;
	meter2 = (double)meter2Percent;

	// Use a polynomial to calculate the real 0-255 based on the logarithmic scale
	meter2position = (255.0/309.0) * (0.0117 * pow(meter2, 2.0) + 1.8663 * meter2 + 3.1713);

	// Set the PWM duty-cycle
	meter2TargetValue = (unsigned char)meter2position;
}

// Check the hardware button and perform debounce
void checkButton(void)
{
	// Debounce the button ON

	// Current button pressed state is 'not pressed', check the physical port pin
	if (BUTTON0 == 0 && buttonPressed == 0) // is pressed?
	{
		// Physical button is pressed
		buttonDebounceCounter++;
		if (buttonDebounceCounter == 5) buttonPressed = 1;
	}
	else
	{
		buttonDebounceCounter = 0;
	}	
}	

// Initialise system
static void InitialiseSystem(void)
{
	// General PIC initialisation ---------------------------------------------
	
	// Configure ports as inputs (1) or outputs(0)
	TRISA = 0b00000000;
	TRISB = 0b00000001;
	TRISC = 0b00000000;

	// Clear/Set ports
	PORTA = 0b00000111;
	PORTB = 0b00000000;
	PORTC = 0b00000000;

	// Turn off analogue ports
	ADCON1 = 0x0F;
	
	// Enable weak pull ups on PORTB (required for the hardware button)
	RBPU = 0;
	
	// Timer0 and interupt initialisation -------------------------------------
	
	// Enable interrupts with priority
	IPEN = 1;

	// Set up timer0
	
	// LED cycle speed is 60Hz with 128 resolution steps (16,667 mS per cycle
	// with 16.67 mS per resolution step)
	
	TMR0IP = 0;			// Set timer0 interrupt to low priority
	TMR0IF = 0;			// Clear the timer0 interrupt flag
	TMR0L = 255 - 98;	// Reset the timer0 counter

	T0CON = 0b11000011; // Timer0 on, 8-bit and 1:16 prescaler
	TMR0IE = 1;			// Enable the timer0 interrupt

	// Enable interrupts
	GIEH = 1;			// Global enable all high priority interrupts
	GIEL = 1;			// Global enable all low priority interrupts
	
	// PWM modules set up -----------------------------------------------------
	
	// Set up timer2 for PWM
	PR2 = 0b11111111;
	T2CON = 0b00000111;

	// Configure PWM module 1 at 0% duty-cycle
	CCPR1L = 0b00000000;
	CCP1CON = 0b00111100;

	// Configure PWM module 2 at 0% duty-cycle
	CCPR2L = 0b00000000;
	CCP2CON = 0b0011110;

	// USB set up -------------------------------------------------------------

	// USB stack and parameter initialisation
    #if defined(USE_SELF_POWER_SENSE_IO)
    tris_self_power = INPUT_PIN;
    #endif
    
    // Initialize the variable holding the handle for the last
    // transmission
    USBOutHandle = 0;
    USBInHandle = 0;
    
	// Call the stack initialisation functions
    USBDeviceInit();
}

// Process input and output
void ProcessIO(void)
{   
    // If we are not in the configured state just return
    if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1)) return;

	// Check the status of the hardware button
	checkButton();

	// Check if data was received from the host.
    if(!HIDRxHandleBusy(USBOutHandle))
    {   
        switch(ReceivedDataBuffer[0])
		{
            case 0x80:
				// Command 0x80: Update the device
				//			Byte[0] - The command 0x80
				//			Byte[1] - The CPU meter percentage		(0-100) %
				//			Byte[2] - The HDD meter percentage		(0-100) %
				//			Byte[3] - The LED red value				(0-255) R
				//			Byte[4] - The LED green value			(0-255) G
				//			Byte[5] - The LED blue value			(0-255) B
				//			Byte[6] - Detach USB on button flag     (0-1) false, true
				
				// Update the meters
				setMeter(ReceivedDataBuffer[1], ReceivedDataBuffer[2],
					meter1SensitivitySetting, meter2SensitivitySetting);
				
				// Update the LED
				setRgbLedColour(ReceivedDataBuffer[3], ReceivedDataBuffer[4], ReceivedDataBuffer[5]);
				
				// Check for soft USB detach command
				if (ReceivedDataBuffer[6] == 1 && buttonPressed == 1) softUsbDetach = 1;
				
				// Now send back the button state to the host
				if (buttonPressed == 1 && ReceivedDataBuffer[6] == 0)
				{
					ToSendDataBuffer[0] = 0x80;
					ToSendDataBuffer[1] = buttonPressed;
					buttonPressed = 0;
				}
				else
				{
					ToSendDataBuffer[0] = 0x80;
					ToSendDataBuffer[1] = 0;
				}
				
				// Transmit the response to the host
                if(!HIDTxHandleBusy(USBInHandle))
				{
					USBInHandle = HIDTxPacket(HID_EP,(BYTE*)&ToSendDataBuffer[0],64);
				}
				
                break;
                
            case 0x81:
            	// Command 0x81: Update the device settings
				//			Byte[0] - The command 0x81
				//			Byte[1] - The meter1 sensitivity			(0-100) %
				//			Byte[2] - The meter2 sensitivity			(0-100) %
				//			Byte[3] - The LED fade on rate low-byte		(0-255) (microseconds)
				//			Byte[4] - The LED fade on rate high-byte	(0-255) (microseconds)
				//			Byte[5] - The LED fade off rate low-byte	(0-255) (microseconds)
				//			Byte[6] - The LED fade off rate high-byte	(0-255) (microseconds)
				
				meter1SensitivitySetting = ReceivedDataBuffer[1];
				meter2SensitivitySetting = ReceivedDataBuffer[2];
				ledFadeOnRateSetting = (ReceivedDataBuffer[4] << 8) + ReceivedDataBuffer[3];
				ledFadeOffRateSetting = (ReceivedDataBuffer[6] << 8) + ReceivedDataBuffer[5];
            	break;
        }
        // Re-arm the OUT endpoint for the next packet
        USBOutHandle = HIDRxPacket(HID_EP,(BYTE*)&ReceivedDataBuffer,64);
    }  
}

// Main function
void main(void)
{
	// Note:  The USB soft-detach mode allows the user to detach the USB
	// device by pressing the hardware button, and then reattach it by
	// pressing again.  This is useful if the device is attached to a KVM
	// and you want to temporarily disable the USB before switching to another
	// machine.
	while(1)
	{
		// If we are not detached initialise and start the USB device
		if (softUsbDetach == 0) InitialiseSystem();
		
	    while(softUsbDetach == 0)
	    {
			// Check bus status and service USB interrupts.
	        USBDeviceTasks();
	
			// Application-specific tasks.
	        ProcessIO();        
	    }
	    
	    // We are in USB soft-detach mode
	    
	    // Reset the meters and the LED to 'off'
		setMeter(0, 0, meter1SensitivitySetting, meter2SensitivitySetting);
		setRgbLedColour(0, 0, 0);
	    
	    // Disable the USB device
		UCON = 0b00000000;
		UCFG = 0b00001000;
		
		// Wait for a button press before re-attaching the USB device
		buttonPressed = 0;
		for (int debounce = 0; debounce < 10; debounce++)
			while (BUTTON0 == 0); // Wait for the button to be released
		
		while (buttonPressed == 0) checkButton();
		
		// Enable the USB device
		UCON = 0b00000000;
		UCFG = 0b00000000;
		
		buttonPressed = 0;
		softUsbDetach = 0;
 	}	   
}

#endif
